#ifndef SIMULATED_ANNEALING
#define SIMULATED_ANNEALING 1

#include <time.h>
#include <iostream>

#include <vector>
#include <algorithm>
#include <boost/foreach.hpp>

#include "optimize.hpp"
#include "util.h"

namespace Optimization{


  typedef std::pair<Value, unsigned int> pair_score_id;
  bool sort_pred_pair_score_id(const pair_score_id& a, 
			       const pair_score_id& b)
  { return a.first < b.first; }


  template<typename CostFunction>
  class GeneticProgramming{

  public: 

    typedef std::vector<Vector> Population;

    //
    //  SimulatedAnnealing
    //
    //   @param Domain to optimize
    //   @param step (the step for moving an index)

    GeneticProgramming(Domain & domain, 
		       Value size_population=30,
		       unsigned int step=1,
		       Value p_mutation=0.2,
		       Value p_elite=0.2,
		       unsigned int num_iter=100) : 
      domain_(domain), size_domain_(domain.size()),
      size_population_(size_population), 
      step_(step), p_mutation_(p_mutation), 
      p_elite_(p_elite), num_iter_(num_iter){};
    
    void optimize(){

      unsigned int i, j, num_elite, iteration, mutant, crossover1, crossover2;
      Value r;
      
      Population population;      // population == n Vectors

      if (size_domain_ == 0){
	std::cout << "Size domain is 0, exiting" << std::endl;
      }

      srand( time(NULL) );        // init the random generator

      //
      // Random init the population of vectors within the domain
      //
      //      std::cout << "Population" << std::endl;
      for (i=0 ; i < size_population_; i++){
	
	Vector vec(size_domain_);
	population.push_back(vec);
	
	for (j=0 ; j < size_domain_; j++)
	  population[i][j] = 
	    int(random_in_interval(domain_[j].first, domain_[j].second));
	
	//std::cout << "\tInitial vector" << population[i] << std::endl;
      }
      
      num_elite = int(p_elite_ * size_population_);  // number topelite
      
      // 
      // Start main genetic computation
      //
      iteration = 0;                             // current iteration
      while (iteration < num_iter_){  

	std::vector<pair_score_id> scores;         // score of population

	std::cout << "Iteration = " << iteration << std::endl;
	// compute the score of each vector in population
	for (i=0 ; i < size_population_; i++)
	  scores.push_back(pair_score_id(CostFunction()(population[i]),i));  

	// sort the score
	std::sort(scores.begin(), scores.end(), sort_pred_pair_score_id);
	
	// take the elite
	for (i=0; i < num_elite; i++){
	
	  std::cout << "\telite solutions, so far" << population[scores[i].second]  
		    << "cost=" << CostFunction()(population[scores[i].second]) << std::endl;
	  //
	  // Try a mutation
	  // 
	  if ((r = random_in_interval(0,1)) < p_mutation_){

	    //  std::cout << "Mutation (r=" << r 
	    //      << "<p=" << p_mutation_ << ")" << std::endl;

	    //
	    // Select one random in the elite 
	    //   Mutate it and add to the population 
	    //
	    mutant = int(random_in_interval(0, num_elite));
	    //std::cout << " \tmutant is = " << scores[mutant].second << "  ";
	    //std::cout << population[scores[mutant].second] << std::endl;
	    population.push_back(mutate(population[scores[mutant].second]));
	    
	  } else {	    
	    //
	    // Try cross-over
	    // 
	    // std::cout << "Crossover (r=" << r << std::endl;
	    crossover1 = int(random_in_interval(0, num_elite));
	    crossover2 = int(random_in_interval(0, num_elite));
	    //std::cout << " \tcross-over1 is=" << scores[crossover1].second << "  ";
	    //std::cout << population[scores[crossover1].second] << std::endl;
	    //std::cout << " \tcross-over2 is=" << scores[crossover2].second << "  ";
	    //std::cout << population[scores[crossover2].second] << std::endl;

	    population.push_back(crossover(population[scores[crossover1].second],
					   population[scores[crossover2].second])
				 );
	  }
	  
	  size_population_ = population.size();  // population may be increased
	}
	iteration++;
      }// while         
    };
    

  private:
    Domain domain_;                 // range for variables
    std::size_t size_domain_;       // number of variables
    unsigned int size_population_;  // size of the population
    unsigned int step_;   
    Value p_mutation_;              // probability of mutation
    Value p_elite_;                 // % of elite
    unsigned int num_iter_;         // # of iteration
 


    // Mutate a vector
    //
    //  @param vector to mutate
    //  @return vector mutated 
    //
    inline Vector mutate(Vector& v){

      Vector v2(v);
      Value r;

      //select and index
      unsigned int i = int(random_in_interval(0, size_domain_));
      
      if ( ( (r=random_in_interval(0, 1)) < 0.5) &&  // direction
	   v[i] > domain_[i].first + step_)          // within bound
	
	v2[i]-=step_;                                // decrease i-th
      
      else if ( v[i] > domain_[i].first - step_)     // within bound
	
	v2[i]+=step_;                                 // increase i-th

      //      std::cout << "\t\t Mutated vector" << v2 << std::endl;
      return v2;
    }

    
    // 
    // Crossover
    //
    // @param v1 vector to crossover
    // @param v2 vector to crossover
    // 
    inline Vector crossover(Vector& v1, Vector& v2){

      Vector v3;
      //select and index
      unsigned int i = int(random_in_interval(0, size_domain_));

      for (unsigned j = 0; j < i; j++)
	v3.push_back(v1[j]);           // copy from v1

      for (unsigned j = i; j < size_domain_; j++)
	v3.push_back(v2[j]);           // copy from v2

      //std::cout << "\t\t Crossover vector" << v3 << std::endl;
      return v3;
    }

  };

};

#endif
